local UH = UnderHood
local UHB = UnderHood_Buffs
local L = LibStub( "AceLocale-3.0" ):GetLocale( "UnderHood" )

local BuffFrame = {}
BuffFrame.__index = BuffFrame
UHB.FramePrototype = BuffFrame

local validTypes = {
	buffs = L["Buffs"],
	debuffs = L["Debuffs"],
}

local validLayouts = {
	LTD = L["Left, then Down"],
	RTD = L["Right, then Down"],
	LTU = L["Left, then Up"],
	RTU = L["Right, then Up"],
	DTL = L["Down, then Left"],
	DTR = L["Down, then Right"],
	UTL = L["Up, then Left"],
	UTR = L["Up, then Right"],
}

local layouts = {
	LTD = function( button, buffFrame, index )
		local columns = buffFrame.config.columns
		local row = math.floor( (index - 1) / columns ) + 1
		local column = ((index - 1) % columns) + 1

	    if column == 1 then
			if row == 1 then
			    button:SetPoint( "TOPRIGHT", buffFrame.frame, "TOPRIGHT" )
			else
			    button:SetPoint( "TOP", buffFrame.buttons[index - columns], "BOTTOM", 0, -buffFrame.config.vSpace )
			end
	    else
		    button:SetPoint( "RIGHT", buffFrame.buttons[index - 1], "LEFT", -buffFrame.config.hSpace, 0 )
	    end
	end,
	RTD = function( button, buffFrame, index )
		local columns = buffFrame.config.columns
		local row = math.floor( (index - 1) / columns ) + 1
		local column = ((index - 1) % columns) + 1

	    if column == 1 then
			if row == 1 then
			    button:SetPoint( "TOPLEFT", buffFrame.frame, "TOPLEFT" )
			else
			    button:SetPoint( "TOP", buffFrame.buttons[index - columns], "BOTTOM", 0, -buffFrame.config.vSpace )
			end
	    else
		    button:SetPoint( "LEFT", buffFrame.buttons[index - 1], "RIGHT", buffFrame.config.hSpace, 0 )
	    end
	end,
	LTU = function( button, buffFrame, index )
		local columns = buffFrame.config.columns
		local row = math.floor( (index - 1) / columns ) + 1
		local column = ((index - 1) % columns) + 1

	    if column == 1 then
			if row == 1 then
			    button:SetPoint( "BOTTOMRIGHT", buffFrame.frame, "BOTTOMRIGHT" )
			else
			    button:SetPoint( "BOTTOM", buffFrame.buttons[index - columns], "TOP", 0, buffFrame.config.vSpace )
			end
	    else
		    button:SetPoint( "RIGHT", buffFrame.buttons[index - 1], "LEFT", -buffFrame.config.hSpace, 0 )
	    end
	end,
	RTU = function( button, buffFrame, index )
		local columns = buffFrame.config.columns
		local row = math.floor( (index - 1) / columns ) + 1
		local column = ((index - 1) % columns) + 1

	    if column == 1 then
			if row == 1 then
			    button:SetPoint( "BOTTOMLEFT", buffFrame.frame, "BOTTOMLEFT" )
			else
			    button:SetPoint( "BOTTOM", buffFrame.buttons[index - columns], "TOP", 0, buffFrame.config.vSpace )
			end
	    else
		    button:SetPoint( "LEFT", buffFrame.buttons[index - 1], "RIGHT", buffFrame.config.hSpace, 0 )
	    end
	end,
	DTL = function( button, buffFrame, index )
		local rows = buffFrame.config.rows
		local row = ((index - 1) % rows) + 1
		local column = math.floor( (index - 1) / rows ) + 1

		if row == 1 then
		    if column == 1 then
		        button:SetPoint( "TOPRIGHT", buffFrame.frame, "TOPRIGHT" )
		    else
		        button:SetPoint( "RIGHT", buffFrame.buttons[index - buffFrame.config.rows], "LEFT", -buffFrame.config.hSpace, 0 )
		    end
		else
		    button:SetPoint( "TOP", buffFrame.buttons[index - 1], "BOTTOM", 0, -buffFrame.config.vSpace )
		end
	end,
	DTR = function( button, buffFrame, index )
		local rows = buffFrame.config.rows
		local row = ((index - 1) % rows) + 1
		local column = math.floor( (index - 1) / rows ) + 1

		if row == 1 then
		    if column == 1 then
		        button:SetPoint( "TOPLEFT", buffFrame.frame, "TOPLEFT" )
		    else
		        button:SetPoint( "LEFT", buffFrame.buttons[index - buffFrame.config.rows], "RIGHT", buffFrame.config.hSpace, 0 )
		    end
		else
		    button:SetPoint( "TOP", buffFrame.buttons[index - 1], "BOTTOM", 0, -buffFrame.config.vSpace )
		end
	end,
	UTL = function( button, buffFrame, index )
		local rows = buffFrame.config.rows
		local row = ((index - 1) % rows) + 1
		local column = math.floor( (index - 1) / rows ) + 1

		if row == 1 then
		    if column == 1 then
		        button:SetPoint( "BOTTOMRIGHT", buffFrame.frame, "BOTTOMRIGHT" )
		    else
		        button:SetPoint( "RIGHT", buffFrame.buttons[index - buffFrame.config.rows], "LEFT", -buffFrame.config.hSpace, 0 )
		    end
		else
		    button:SetPoint( "BOTTOM", buffFrame.buttons[index - 1], "TOP", 0, buffFrame.config.vSpace )
		end
	end,
	UTR = function( button, buffFrame, index )
		local rows = buffFrame.config.rows
		local row = ((index - 1) % rows) + 1
		local column = math.floor( (index - 1) / rows ) + 1

		if row == 1 then
		    if column == 1 then
		        button:SetPoint( "BOTTOMLEFT", buffFrame.frame, "BOTTOMLEFT" )
		    else
		        button:SetPoint( "LEFT", buffFrame.buttons[index - buffFrame.config.rows], "RIGHT", buffFrame.config.hSpace, 0 )
		    end
		else
		    button:SetPoint( "BOTTOM", buffFrame.buttons[index - 1], "TOP", 0, buffFrame.config.vSpace )
		end
	end,
}

function BuffFrame:Create( module, config )
	local self = {}
	setmetatable( self, BuffFrame )

	LibStub( "AceTimer-3.0" ):Embed( self )
	LibStub( "AceEvent-3.0" ):Embed( self )
	
	self.module = module
	self.config = config
	
	self.frame = UH:CreateFrame( "Frame" ) -- parented to UnderHoodParentFrame
	self.frame:SetScale( self.config.scale )
	self.frame:SetFrameStrata( self.config.strata )
	self.frame:SetFrameLevel( self.config.level )
	self.frame:SetPoint( "CENTER", UIParent, "CENTER", self.config.x, self.config.y )
	--self.frame:SetAlpha( UH:GetAlpha() )
	
	self:ResizeFrame()
	self.frame:Show()

	self.buttons = {}
	
	for i = 1, 40 do
	    local button = UH:CreateFrame( "Button", self.frame )

		button:SetWidth( self.config.size )
		button:SetHeight( self.config.size )
		button:SetFrameLevel( self.config.level + 1 )
		
		button:SetID( i )
		
		button:SetBackdrop( {
		    bgFile = "Interface\\Buttons\\WHITE8X8", tile = true, tileSize = 8,
		    edgeFile = "Interface\\Buttons\\WHITE8X8", edgeSize = 1,
		    insets = { left = 1, right = 1, top = 1, bottom = 1 },
		} )
		
		button:SetBackdropColor( 0, 0, 0, 0 )
		button:SetBackdropBorderColor( 0, 0, 0, 1 )

		button.Icon = UH:CreateFrame( "Texture", button, "BACKGROUND" )
		button.Icon:SetTexture( "Interface\\Icons\\INV_Misc_Ear_Human_02" )
		button.Icon:SetTexCoord( 0.07, 0.93, 0.07, 0.93 )
		button.Icon:SetPoint( "TOPLEFT", button, "TOPLEFT", 1, -1 )
		button.Icon:SetPoint( "BOTTOMRIGHT", button, "BOTTOMRIGHT", -1, 1 )
		button.Icon:Show()
		
		--[[
		button.Border = UH:CreateFrame( "Texture", button, "BORDER" )
		button.Border:SetTexture( "Interface\\Buttons\\UI-Debuff-Overlays" )
		button.Border:SetTexCoord( 0.296875, 0.5703125, 0, 0.515625 )
		button.Border:SetWidth( self.config.size + 1)
		button.Border:SetHeight( self.config.size + 1 )
		button.Border:SetPoint( "CENTER", button, "CENTER" )
		button.Border:Hide()
		]]
		
		button.Count = UH:CreateFrame( "FontString", button, "OVERLAY", "NumberFontNormalSmall" )
		button.Count:SetPoint( "BOTTOMRIGHT", button, "BOTTOMRIGHT", -2, 1 )
		button.Count:Show()

		button.cooldown = UH:CreateFrame( "Cooldown", button )
		button.cooldown:SetReverse( true )
		button.cooldown:SetPoint( "TOPLEFT", button, "TOPLEFT", 1, -1 )
		button.cooldown:SetPoint( "BOTTOMRIGHT", button, "BOTTOMRIGHT", -1, 1 )
		button.cooldown:Hide()
		
		button.cooldownText = UH:CreateFrame( "FontString", button.cooldown, "OVERLAY", "NumberFontNormalSmall" )
		button.cooldownText:SetTextColor( 1, 1, 1, 1 )
		button.cooldownText:SetShadowColor( 0, 0, 0, 1 )
		button.cooldownText:SetShadowOffset( 0.8, -0.8 )
		button.cooldownText:SetPoint( "TOPLEFT", button, "TOPLEFT", 1, -1 )
		button.cooldownText:SetPoint( "BOTTOMRIGHT", button, "BOTTOMRIGHT", -1, 1 )
		button.cooldownText:Hide()
		
		button.owner = self
		
		button:SetScript( "OnEnter", function()
			if this.owner.config.tooltips == "always"
				or (this.owner.config.tooltips == "ooc" and not( InCombatLockdown() )) then
				if not this:IsVisible() then return end
				
				GameTooltip:SetOwner( this, "ANCHOR_BOTTOMRIGHT" )
				
				if this.owner.config.type == "buffs" then
				    GameTooltip:SetUnitBuff( this.owner.config.unit, this:GetID() )
				else
				    GameTooltip:SetUnitDebuff( this.owner.config.unit, this:GetID() )
				end
				
				GameTooltip:Show()
		    end
		end )
		
		button:SetScript( "OnLeave", function()
		    if GameTooltip:IsOwned( this ) then
		        GameTooltip:Hide()
			end
		end )
		
		table.insert( self.buttons, button )
	end

	self:PerformLayout()
	self:UpdateButtons()

	self.frame:SetAttribute( "unit", self.config.unit )
	self.frame:SetAttribute( "show", "unit" )
	self.frame:SetAttribute( "useunit", self.config.unit )
	
	RegisterUnitWatch( self.frame )

	self:RegisterEvent( "UNIT_AURA" )
	self:RegisterEvent( "PLAYER_TARGET_CHANGED" )
	self:RegisterEvent( "PLAYER_FOCUS_CHANGED" )

	--if self.config.unit == "targettarget" then
	--    self:ScheduleRepeatingTimer( "UpdateButtons", 0.1 )
	--end
	
	self:ScheduleRepeatingTimer( "ProcessUpdate", 0.1 )
	
	self:SetupOptions()

	return self
end

function BuffFrame:Destroy()
	self:CancelAllTimers()
	
	for _, button in ipairs( self.buttons ) do
	    UH:ReleaseFrame( button.cooldownText )
	    UH:ReleaseFrame( button.cooldown )
	    UH:ReleaseFrame( button.Count )
	    UH:ReleaseFrame( button.Border )
	    UH:ReleaseFrame( button.Icon )
	    
	    button.cooldownText = nil
	    button.cooldown = nil
	    button.Count = nil
	    button.Border = nil
	    button.Icon = nil
	    
	    button.Update = nil
	    button.owner = nil
	    
	    UH:ReleaseFrame( button )
	end
	
	self.buttons = nil
	
	if self.frame then
		self.frame:SetAttribute( "unit", nil )
		self.frame:SetAttribute( "show", nil )
		self.frame:SetAttribute( "useunit", nil )
		
		UnregisterUnitWatch( self.frame )
		
		UH:ReleaseFrame( self.frame )
		self.frame = nil
	end
	
	self:UnregisterAllEvents()
	
	self.frame = nil
end

local function getUnitAura( unit, index, isBuff )
	local name, rank, iconTexture, count, debuffType, duration, timeLeft
	
	if isBuff then
	    name, rank, iconTexture, count, duration, timeLeft = UnitBuff( unit, index )
	else
	    name, rank, iconTexture, count, debuffType, duration, timeLeft = UnitDebuff( unit, index )
	end
	
	return name, iconTexture, count, debuffType, duration, timeLeft
end

-- From PitBull_Aura, thanks once again :)

local getPlayerAuraDuration
do
	local buffData = {}
	local debuffData = {}
	
	function getPlayerAuraDuration( isBuff, key, providedDuration )
		local data = isBuff and buffData or debuffData
		local duration = data[key]
		if not duration then
			for i = 1, 40 do
				local name, _, time
				if isBuff then
					name, _, _, _, time = UnitBuff("player", i)
				else
					name, _, _, _, _, time = UnitDebuff("player", i)
				end
				if not name then
					break
				end
				if name == key then
					duration = time
					break
				end
			end
		end
		if not duration or duration < providedDuration then
			duration = providedDuration
		end
		data[key] = duration
		return duration
	end
end

function BuffFrame:FilterAura( isBuff, name, debuffType )
	if self.config.filterDispelable then
		if not( debuffType ) or not( self.module.DispelableDebuffs[debuffType] ) then
			return
		end
	end
	
	return true
end

function BuffFrame:CollectAuras()
	local uid = self.config.unit
	local isBuff = self.config.type == "buffs"
	local auras = UH:NewList()
	local auraCount = 0
	local maxAuras = math.min( self.config.columns * self.config.rows, 40 )

	if uid == "player" then
	    local id
	    local i = 1
	    
	    while true do
	        if auraCount > maxAuras then break end
	        
			id = GetPlayerBuff( i, isBuff and "HELPFUL" or "HARMFUL" )
			if id == 0 then break end
			
			local name, iconTexture, count, debuffType, timeLeft = GetPlayerBuffName( id ), GetPlayerBuffTexture( id ), GetPlayerBuffApplications( id ), GetPlayerBuffDispelType( id ), GetPlayerBuffTimeLeft( id )
			local duration = getPlayerAuraDuration( isBuff, name, timeLeft )

			if self:FilterAura( isBuff, name, debuffType ) then
				auras[#auras+1] = UH:NewList( name, iconTexture, count, debuffType, duration, timeLeft )
				auraCount = auraCount + 1
			end
			
			i = i + 1
		end
	else
	    local id = 1
	    
	    while true do
	        if auraCount > maxAuras then break end

			local name, iconTexture, count, debuffType, duration, timeLeft = getUnitAura( uid, id, isBuff )
			
			if not name then break end

			if self:FilterAura( isBuff, name, debuffType ) then
				auras[#auras+1] = UH:NewList( name, iconTexture, count, debuffType, duration, timeLeft )
				auraCount = auraCount + 1
			end

			id = id + 1
	    end
	end
	
	return auras
end

function BuffFrame:UpdateButtons()
	local auras = self:CollectAuras()
	local maxButtons = math.min( self.config.columns * self.config.rows, 40 )
	local isBuff = self.config.type == "buffs"
	
	for i, button in ipairs( self.buttons ) do
		if i > maxButtons and button:IsVisible() then
		    button:Hide()
		else
		    if i > #auras then
				if self:InConfigMode() then
				    button.Icon:SetTexture( "Interface\\Icons\\INV_Misc_Ear_Human_02" )
				    button.Count:SetText( i )
				    button.Count:Show()

				    button:Show()
				elseif button:IsVisible() then
				    button:Hide()
				end
		    else
		    	local name, texture, count, debuffType, duration, timeLeft = unpack( auras[i] )

		        button.Icon:SetTexture( texture )

		        if count and count > 1 then
		            button.Count:SetText( count )
		            button.Count:Show()
		        else
					button.Count:Hide()
		        end
		        
		        if duration and duration > 0 then
		            button.cooldown:Show()
		            button.cooldown:SetCooldown( GetTime()+timeLeft-duration, duration )
				else
				    button.cooldown:Hide()
		        end

				if isBuff or not self.config.colorDebuffBorders then
					button:SetBackdropBorderColor( 0, 0, 0, 1 )
				else
		        	local color = DebuffTypeColor[debuffType or "none"]

					button:SetBackdropBorderColor( color.r, color.g, color.b, 1 )
				end
				
		        button:Show()
			end
		end
	end
	
	for i, v in ipairs( auras ) do
	    auras[i] = UH:ReleaseTable( auras[i] )
	end
	
	auras = UH:ReleaseTable( auras )
end

function BuffFrame:ProcessUpdate()
	if self.needUpdate or self.config.unit == "targettarget" then
	    self.needUpdate = nil
	    self:UpdateButtons()
	end
end

function BuffFrame:UNIT_AURA( eventName, unitId )
	if unitId == self.config.unit then
	    self.needUpdate = true
	end
end

function BuffFrame:PLAYER_TARGET_CHANGED()
	if self.config.unit == "target" or self.config.unit == "targettarget" then
	    self.needUpdate = true
	end
end

function BuffFrame:PLAYER_FOCUS_CHANGED()
	if self.config.unit == "focus" then
	    self.needUpdate = true
	end
end

function BuffFrame:ResizeFrame()
	local width = (self.config.columns * self.config.size) + (self.config.columns - 1)*self.config.hSpace
	local height = (self.config.rows * self.config.size) + (self.config.rows - 1)*self.config.vSpace

	self.frame:SetWidth( width )
	self.frame:SetHeight( height )
end

function BuffFrame:PerformLayout()
	local maxButtons = math.min( self.config.columns * self.config.rows, 40 )
	local layout = layouts[self.config.layout]
	
	for i, button in ipairs( self.buttons ) do
		button:SetWidth( self.config.size )
		button:SetHeight( self.config.size )
	    button:ClearAllPoints()
	    layout( button, self, i )
	end
end

function BuffFrame:InConfigMode()
	return UH:InConfigMode()
end

function BuffFrame:SetConfigMode( mode )
	if mode then
		UnregisterUnitWatch( self.frame )
		
		self.frame:Show()
	else
		if UnitExists( self.config.unit ) then
			self.frame:Show()
		else
			self.frame:Hide()
		end
		
		RegisterUnitWatch( self.frame )
	end
	
	self.needUpdate = true
end

function BuffFrame:ChangeUnit( unit )
	if unit and unit == self.config.unit then return end
	
	UnregisterUnitWatch( self.frame )
	
	self.config.unit = unit

	self.frame:SetAttribute( "unit", unit )
	self.frame:SetAttribute( "shown", "unit" )
	self.frame:SetAttribute( "useunit", unit )

	if not self:InConfigMode() then
		RegisterUnitWatch( self.frame )
	else
		if UnitExists( self.config.unit ) then
			self.frame:Show()
		else
			self.frame:Hide()
		end
	end
	
	self.needUpdate = true
	
	--self:UpdateButtons()
end

function BuffFrame:SetupOptions()
	self.options = {
		name = function() return self.config.name end,
		type = "group",
		childGroups = "tab",
		args = {
			unitAndType = {
				name = L["Unit and Type"],
				type = "group",
				--inline = true,
				order = 100,
				args = {
					unit = {
						name = L["Unit"],
						type = "select",
						order = 1,
						values = UH.OptionsHelper.Unit,
						get = function() return self.config.unit end,
						set = function( info, value ) self:ChangeUnit( value ) end,
					},
					type = {
						name = L["Type"],
						type = "select",
						order = 2,
						values = validTypes,
						get = function( info ) return self.config.type end,
						set = function( info, value ) self.config.type = value; self.needUpdate = true end,
					},
					tooltips = {
					    name = L["Show Tooltips"],
					    type = "select",
					    order = 3,
					    values = UH.OptionsHelper.TooltipVisibility,
					    get = function() return self.config.tooltips end,
					    set = function( info, value ) self.config.tooltips = value end,
					},
					header1 = {
						name = L["Debuff options"],
						type = "header",
						order = 4,
						hidden = function() return self.config.type ~= "debuffs" end,
					},
					colorDebuffBorders = {
						name = L["Color borders"],
						desc = L["Color borders using debuff type color"],
						type = "toggle",
						order = 5,
						hidden = function() return self.config.type ~= "debuffs" end,
						get = function() return self.config.colorDebuffBorders end,
						set = function( info, value ) self.config.colorDebuffBorders = true; self.needUpdate = true end,
					},
					debuffFilter = {
					    name = L["Dispelable only"],
						desc = L["Show debuffs you can dispel only"],
						type = "toggle",
						order = 6,
						hidden = function() return self.config.type ~= "debuffs" end,
						get = function() return self.config.filterDispelable end,
						set = function( info, value ) self.config.filterDispelable = value; self.needUpdate = true end,
					},
				},
			},
			positionAndSize = {
				name = L["Position and Size"],
				type = "group",
				--inline = true,
				order = 102,
				args = {
					x = {
						name = L["X"],
						type = "range",
						order = 1,
						min = -800,
						max = 800,
						step = 1,
						bigStep = 1,
						get = function() return self.config.x end,
						set = function( info, value )
							self.config.x = value
							self.frame:ClearAllPoints()
							self.frame:SetPoint( "CENTER", UIParent, "CENTER", self.config.x, self.config.y )
						end,
					},
					y = {
						name = L["Y"],
						type = "range",
						order = 2,
						min = -800,
						max = 800,
						step = 1,
						bigStep = 1,
						get = function() return self.config.y end,
						set = function( info, value )
							self.config.y = value
							self.frame:ClearAllPoints()
							self.frame:SetPoint( "CENTER", UIParent, "CENTER", self.config.x, self.config.y )
						end,
					},
					layout = {
						name = L["Layout"],
						type = "select",
						order = 3,
						values = validLayouts,
						get = function() return self.config.layout end,
						set = function( info, value ) self.config.layout = value; self:PerformLayout() end,
					},
					size = {
					    name = L["Button size"],
					    type = "range",
					    order = 4,
					    min = 10,
					    max = 48,
					    step = 1,
					    bigStep = 1,
					    get = function() return self.config.size end,
					    set = function( info, value )
					        self.config.size = value
					        self:ResizeFrame()
					        self:PerformLayout()
					    end,
					},
					columns = {
					    name = L["Columns"],
					    type = "range",
					    order = 5,
					    min = 1,
					    max = 40,
					    step = 1,
					    bigStep = 1,
					    get = function() return self.config.columns end,
					    set = function( info, value )
					        self.config.columns = value
					        self:ResizeFrame()
					        self:PerformLayout()
					        self:UpdateButtons()
						end,
					},
					rows = {
					    name = L["Rows"],
					    type = "range",
					    order = 6,
					    min = 1,
					    max = 40,
					    step = 1,
					    bigStep = 1,
					    get = function() return self.config.rows end,
					    set = function( info, value )
					        self.config.rows = value
					        self:ResizeFrame()
					        self:PerformLayout()
					        self:UpdateButtons()
						end,
					},
					hSpace = {
						name = L["Space between columns"],
						type = "range",
						order = 7,
						min = 0,
						max = 40,
						step = 1,
						bigStep = 1,
						get = function() return self.config.hSpace end,
						set = function( info, value )
							self.config.hSpace = value
							self:ResizeFrame()
							self:PerformLayout()
						end,
					},
					vSpace = {
						name = L["Space between rows"],
						type = "range",
						order = 8,
						min = 0,
						max = 40,
						step = 1,
						bigStep = 1,
						get = function() return self.config.vSpace end,
						set = function( info, value )
							self.config.vSpace = value
							self:ResizeFrame()
							self:PerformLayout()
						end,
					},
					scale = {
						name = L["Scale"],
						type = "range",
						order = 9,
						min = 0.01,
						max = 1,
						step = 0.01,
						bigStep = 0.01,
						get = function() return self.config.scale end,
						set = function( info, value )
							self.config.scale = value
							self.frame:SetScale( value )
						end,
					},
				},
			},
			zOrder = {
				name = L["Z-Order"],
				type = "group",
				--inline = true,
				order = 103,
				args = {
					strata = {
						name = L["Strata"],
						type = "select",
						order = 1,
						values = UH.OptionsHelper.Strata,
						get = function() return self.config.strata end,
						set = function( info, value ) self.config.strata = value; self.frame:SetFrameStrata( value ) end,
					},
					level = {
						name = L["Level"],
						type = "range",
						order = 2,
						min = 1,
						max = 10,
						step = 1,
						get = function() return self.config.level end,
						set = function( info, value ) self.config.level = value; self.frame:SetFrameLevel( value ) end,
					},
				},
			},
		},
		plugins = {},
	}
end
